/*
    IHex - Binary to Intel Hex convertor (Version 1.4)
    Copyright (C) 1999 James Carter <james@cs.york.ac.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>

#define USAGE "ihex [-r <record_length>] [-a <address>] " \
              "[-b <bits_per_offset_block>] [-e] [-n] [-z] " \
              "[-o <filename>] [<filename>]"

int
main(int argc, char **argv)
{
  int i;
  unsigned int rlen=32;
  unsigned int address=0;
  unsigned long offset=0;
  unsigned int bpo=16;
  unsigned char no_offset=0;
  unsigned char old_eof=0;
  unsigned char cpm_eof=0;
  unsigned long filelength=0;
  unsigned char buf[256];
  unsigned int bytes;
  unsigned long checksum;
  FILE *output=stdout;
  FILE *input=stdin;

  while ((i=getopt(argc,argv,"enzb:o:r:a:"))!=EOF)
    switch (i) {
    case 'e':
      old_eof=1;
      break;
    case 'n':
      no_offset=1;
      break;
    case 'b':
      bpo=atoi(optarg);
      if (bpo>16 || bpo<4) {
        fprintf(stderr,"%s: bad bits per offset block %s\n",argv[0],optarg);
        exit(1);
      }
      break;
    case 'o':
      output=fopen(optarg,"w");
      if (!output) {
        fprintf(stderr,"%s: cannot open %s for writing\n",argv[0],optarg);
        exit(1);
      }
      break;
    case 'r':
      rlen=atoi(optarg);
      if (rlen>255) {
        fprintf(stderr,"%s: bad record length %s\n",argv[0],optarg);
        exit(1);
      }
      break;
    case 'a':
      address=atoi(optarg);
      break;
    case 'z':
      cpm_eof=1;
      break;
    default:
      fprintf(stderr,"%s\n",USAGE);
      exit(1);
  }
  if (optind<argc) {
    input=fopen(argv[optind],"r");
    if (!input) {
      fprintf(stderr,"%s: cannot open %s for reading\n",argv[0],argv[optind]);
      exit(1);
    }
    if (optind+1<argc) {
      fprintf(stderr,"Too many args\n%s\n",USAGE);
      exit(1);
    }
  }
  if (no_offset) bpo=16;
  while ((bytes=fread(buf,1,rlen,input))>0) {
    if (address/(1<<bpo)) {
      unsigned char tmpbuf[9];

      if (no_offset) {
        fprintf(stderr,"%s: warning: file exceeds 64k limit\n",argv[0]);
        address=0;
      } else {
        offset+=address&~((1<<bpo)-1);
        address=address&((1<<bpo)-1);
        sprintf(tmpbuf,"%04X",offset>>4);
        checksum=4+(offset>>12)%256+(offset>>20)%256;
        fprintf(output,":02000002%s%02X\r\n",tmpbuf,~(checksum-1)%256);
        filelength+=17;
      }
    }
    checksum=bytes%256
             +(bytes>>8)%256
             +address%256
             +(address>>8)%256;
    fprintf(output,":%02X%04X00",bytes,address);
    filelength+=9;
    address+=bytes;
    for (i=0;i<bytes;i++) {
      fprintf(output,"%02X",buf[i]);
      filelength+=2;
      checksum+=buf[i];
    }
    fprintf(output,"%02X\r\n",~(checksum-1)%256);
    filelength+=4;
  }
  if (old_eof) {
    fprintf(output,":0000000000\r\n");
  } else {
    fprintf(output,":00000001FF\r\n");
  }
  if (cpm_eof) fprintf(output,"%c",0x1a);
  filelength+=14;
  while (cpm_eof && filelength%256) {
    fprintf(output,"%c",0x1a);
    filelength++;
  }
  exit(0);
}
